import * as THREE from '../build/three.module.js';
import { DragControls } from './jsm/controls/DragControls.js';

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        this.domWebGL = document.createElement('div');
        document.body.appendChild(this.domWebGL);

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        renderer.shadowMap.enabled = true;
        renderer.shadowMap.type = THREE.PCFShadowMap;
        this.domWebGL.appendChild(renderer.domElement);
        this._renderer = renderer;

        const scene = new THREE.Scene();
        scene.background = new THREE.Color(0xf0f0f0);
        this._scene = scene;

        this._setupCamera();
        this._setupLight();
        this._setupModel();
        this._setupControls();
        
        window.onresize = this.resize.bind(this);

        this.resize();

        this.render();
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(70, window.innerWidth / window.innerHeight, 1, 5000);
        camera.position.z = 1000;
        this._camera = camera;
    }

    _setupControls() {
        this._enableSelection = false;

		const mouse = new THREE.Vector2(), raycaster = new THREE.Raycaster();
        
        const controls = new DragControls([ ... this._objects], this._camera, this._renderer.domElement);
        
        //controls.addEventListener("drag", this.render.bind(this));
        document.addEventListener("click", this._onClick.bind(this));
        document.addEventListener("keydown", this._onKeyDown.bind(this));
        document.addEventListener("keyup", this._onKeyUp.bind(this));

        this._controls = controls;
        this._mouse = mouse;
        this._raycaster = raycaster;
    }

    _onClick(event) {
        event.preventDefault();

        if(this._enableSelection) {
            const draggableObjects = this._controls.getObjects();
            draggableObjects.length = 0;

            this._mouse.x = (event.clientX / window.innerWidth) * 2 - 1;
            this._mouse.y = -(event.clientY / window.innerHeight) * 2 + 1;
            this._raycaster.setFromCamera(this._mouse, this._camera);

            const intersections = this._raycaster.intersectObjects(this._objects, true);

            if(intersections.length > 0) {
                const object = intersections[0].object;

                if(this._group.children.includes(object)) {
                    object.material.emissive.set(0x000000);
                    this._scene.attach(object);
                } else {
                    object.material.emissive.set(0xaaaaaa);
                    this._group.attach(object);
                }

                this._controls.transformGroup = true;
                draggableObjects.push(this._group);
            }

            if(this._group.children.length === 0) {
                this._controls.transformGroup = false;
                draggableObjects.push( ...this._objects);
            }
        }
    }

    _onKeyDown(event) {
        this._enableSelection = (event.keyCode === 16) ? true : false; // 16 = Shift Key
    }

    _onKeyUp(event) {
        this._enableSelection = false;
    }

    _setupModel() {
        const group = new THREE.Group();
        this._scene.add(group);

        const objects = [];

        const geometry = new THREE.BoxGeometry(40, 40, 40);
        for(let i=0; i<200; i++) {
            const object = new THREE.Mesh(geometry, new THREE.MeshLambertMaterial({color: Math.random()*0xffffff}));

            object.position.x = Math.random() * 1000 - 500;
            object.position.y = Math.random() * 600 - 300;
            object.position.z = Math.random() * 800 - 400;

            object.rotation.x = Math.random() * 2 * Math.PI;
            object.rotation.y = Math.random() * 2 * Math.PI;
            object.rotation.z = Math.random() * 2 * Math.PI;

            object.scale.x = Math.random() * 2 + 1;
            object.scale.y = Math.random() * 2 + 1;
            object.scale.z = Math.random() * 2 + 1;

            object.castShadow = true;
            object.receiveShadow = true;

            this._scene.add(object);
            objects.push(object);            
        }

        this._group = group;
        this._objects = objects;
    }

    _setupLight() {
        this._scene.add(new THREE.AmbientLight(0x505050));

        const light = new THREE.SpotLight(0xffffff, 1.5);
        light.position.set(0, 500, 2000);
        light.angle = Math.PI / 9;
        light.castShadow = true;
        light.shadow.camera.near = 1000;
        light.shadow.camera.far = 4000;
        light.shadow.mapSize.width = 1024;
        light.shadow.mapSize.height = 1024;
        this._scene.add(light);
    }

    update() {
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._camera);
    }

    resize() {
        this._camera.aspect = window.innerWidth / window.innerHeight;
        this._camera.updateProjectionMatrix();

        this._renderer.setPixelRatio(window.devicePixelRatio);
        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }
}

window.onload = function () {
    (new App()).render();
}